#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import sys
import os
import os.path
import argparse

from . import app


class ArgParse(argparse.ArgumentParser):
    def __init__(self):
        super(ArgParse, self).__init__(
            description = 'Web file browser'
        )
        self.add_argument('host', nargs='?', default='127.0.0.1', help='address to listen (default: 127.0.0.1)')
        self.add_argument('port', nargs='?', default=8080, type=int, help='port to listen (default: 8080)')
        self.add_argument('--directory', metavar='PATH', type=self._directory, help='base serving directory (default: current path)')
        self.add_argument('--initial', metavar='PATH', type=self._directory, help='initial directory (default: same as --directory)')
        self.add_argument('--removable', metavar='PATH', type=self._directory, help='base directory for remove (default: none)')
        self.add_argument('--upload', metavar='PATH', type=self._directory, help='base directory for upload (default: none)')
        self.add_argument('--debug', action='store_true', help='debug mode')

    def _directory(self, arg):
        if os.path.isdir(arg):
            return os.path.abspath(arg)
        self.error('%s is not a valid directory' % arg)


if __name__ == '__main__':
    cwd = os.path.abspath(os.getcwd())
    args = ArgParse().parse_args(sys.argv[1:])
    app.debug = args.debug
    app.config.update(
        directory_base = args.directory or cwd,
        directory_start = args.initial or args.directory or cwd,
        directory_remove = args.removable if args.removable else None,
        directory_upload = args.upload if args.upload else None,
        )
    app.run(host=os.getenv('IP', args.host), port=args.port)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import sys
import unittest
import re
import os
import os.path
import shutil
import tempfile
import tarfile
import xml.etree.ElementTree as ET
import io
import stat

import flask
import browsepy

PY_LEGACY = browsepy.PY_LEGACY

class Page(object):
    @classmethod
    def itertext(cls, element):
        # TODO(ergoithz) on python2 drop: replace by element.itertext()
        yield element.text or ''
        for child in element:
            for text in cls.itertext(child):
                yield text
            yield child.tail or ''

    @classmethod
    def innerText(cls, element):
        return ''.join(cls.itertext(element))


class ListPage(Page):
    path_strip_re = re.compile('\s+/\s+')
    def __init__(self, path, directories, files, removable, upload):
        self.path = path
        self.directories = directories
        self.files = files
        self.removable = removable
        self.upload = upload


    @classmethod
    def from_source(cls, source):
        html = ET.fromstring(source)
        rows = [
            (
                row[0].attrib.get('class') == 'dir-icon',
                row[1].find('.//a').attrib['href'],
                any(button.attrib.get('class') == 'remove button' for button in row[2].findall('.//a'))
            )
            for row in html.findall('.//table/tbody/tr')
        ]
        return cls(
            cls.path_strip_re.sub('/', cls.innerText(html.find('.//h1'))).strip(),
            [url for isdir, url, removable in rows if isdir],
            [url for isdir, url, removable in rows if not isdir],
            all(removable for isdir, url, removable in rows) if rows else False,
            html.find('.//form//input[@type=\'file\']') is not None
        )


class ConfirmPage(Page):
    def __init__(self, path, back):
        self.path = path
        self.back = back

    @classmethod
    def from_source(cls, source):
        html = ET.fromstring(source)

        return cls(
            cls.innerText(html.find('.//form//strong')).strip(),
            html.find('.//form//a').attrib['href']
        )


class PageException(Exception):
    def __init__(self, status):
        self.status = status


class Page404Exception(PageException):
    pass


class TestApp(unittest.TestCase):
    list_page_class = ListPage
    confirm_page_class = ConfirmPage
    page_exceptions = {
        404: Page404Exception,
        None: PageException
    }

    def setUp(self):
        self.app = browsepy.app
        self.base = tempfile.mkdtemp()
        self.start = os.path.join(self.base, 'start')
        self.remove = os.path.join(self.base, 'remove')
        self.upload = os.path.join(self.base, 'upload')

        os.mkdir(self.start)
        os.mkdir(self.remove)
        os.mkdir(self.upload)

        open(os.path.join(self.start, 'testfile.txt'), 'w').close()
        open(os.path.join(self.remove, 'testfile.txt'), 'w').close()

        self.app.config.update(
            directory_base = self.base,
            directory_start = self.start,
            directory_remove = self.remove,
            directory_upload = self.upload,
            SERVER_NAME = 'test',
        )

        self.base_directories = [
            self.url_for('browse', path='remove'),
            self.url_for('browse', path='start'),
            self.url_for('browse', path='upload'),
            ]
        self.start_files = [self.url_for('open', path='start/testfile.txt')]
        self.remove_files = [self.url_for('open', path='remove/testfile.txt')]
        self.upload_files = []

    def clear(self, path):
        assert path.startswith(self.base + os.sep), 'Cannot clear directories out of base'

        for sub in os.listdir(path):
            sub = os.path.join(path, sub)
            if os.path.isdir(sub):
                shutil.rmtree(sub)
            else:
                os.remove(sub)

    def tearDown(self):
        shutil.rmtree(self.base)

    def get(self, endpoint, **kwargs):
        if endpoint in ('index', 'browse'):
            page_class = self.list_page_class
        elif endpoint == 'remove':
            page_class = self.confirm_page_class
        else:
            page_class = None

        with self.app.test_client() as client:
            response = client.get(self.url_for(endpoint, **kwargs))
            if response.status_code != 200:
                raise self.page_exceptions.get(response.status_code, self.page_exceptions[None])(response.status_code)
            return response.data if page_class is None else page_class.from_source(response.data)

    def post(self, endpoint, **kwargs):
        data = kwargs.pop('data') if 'data' in kwargs else {}
        with self.app.test_client() as client:
            response = client.post(self.url_for(endpoint, **kwargs), data=data, follow_redirects=True)
            if response.status_code != 200:
                raise self.page_exceptions.get(response.status_code, self.page_exceptions[None])(response.status_code)
            return self.list_page_class.from_source(response.data)

    def url_for(self, endpoint, **kwargs):
        with self.app.app_context():
            return flask.url_for(endpoint, _external=False, **kwargs)

    def test_index(self):
        page = self.get('index')
        self.assertEqual(page.path, '%s/start' % os.path.basename(self.base))

        self.app.config['directory_start'] = os.path.join(self.base, '..')

        self.assertRaises(
            Page404Exception,
            self.get, 'index'
        )

        self.app.config['directory_start'] = self.start

    def test_browse(self):
        basename = os.path.basename(self.base)
        page = self.get('browse')
        self.assertEqual(page.path, basename)
        self.assertEqual(page.directories, self.base_directories)
        self.assertFalse(page.removable)
        self.assertFalse(page.upload)

        page = self.get('browse', path='start')
        self.assertEqual(page.path, '%s/start' % basename)
        self.assertEqual(page.files, self.start_files)
        self.assertFalse(page.removable)
        self.assertFalse(page.upload)

        page = self.get('browse', path='remove')
        self.assertEqual(page.path, '%s/remove' % basename)
        self.assertEqual(page.files, self.remove_files)
        self.assertTrue(page.removable)
        self.assertFalse(page.upload)

        page = self.get('browse', path='upload')
        self.assertEqual(page.path, '%s/upload' % basename)
        self.assertEqual(page.files, self.upload_files)
        self.assertFalse(page.removable)
        self.assertTrue(page.upload)

        self.assertRaises(
            Page404Exception,
            self.get, 'browse', path='..'
        )

    def test_open(self):
        content = b'hello world'
        with open(os.path.join(self.start, 'testfile3.txt'), 'wb') as f:
            f.write(content)

        data = self.get('open', path='start/testfile3.txt')
        self.assertEqual(data, content)

        self.assertRaises(
            Page404Exception,
            self.get, 'open', path='../shall_not_pass.txt'
        )

    def test_remove(self):
        open(os.path.join(self.remove, 'testfile2.txt'), 'w').close()
        page = self.get('remove', path='remove/testfile2.txt')
        self.assertEqual(page.path, 'remove/testfile2.txt')
        self.assertEqual(page.back, self.url_for('browse', path='remove'))

        basename = os.path.basename(self.base)
        page = self.post('remove', path='remove/testfile2.txt')
        self.assertEqual(page.path, '%s/remove' % basename)
        self.assertEqual(page.files, self.remove_files)

        os.mkdir(os.path.join(self.remove, 'directory'))
        page = self.post('remove', path='remove/directory')
        self.assertEqual(page.path, '%s/remove' % basename)
        self.assertEqual(page.files, self.remove_files)

        self.assertRaises(
            Page404Exception,
            self.get, 'remove', path='start/testfile.txt'
        )

        self.assertRaises(
            Page404Exception,
            self.post, 'remove', path='start/testfile.txt'
        )

        self.app.config['directory_remove'] = None

        self.assertRaises(
            Page404Exception,
            self.get, 'remove', path='remove/testfile.txt'
        )

        self.app.config['directory_remove'] = self.remove

        self.assertRaises(
            Page404Exception,
            self.get, 'remove', path='../shall_not_pass.txt'
        )


    def test_download_file(self):
        binfile = os.path.join(self.base, 'testfile.bin')
        bindata = bytes(range(256))

        with open(binfile, 'wb') as f:
            f.write(bindata)
        data = self.get('download_file', path='testfile.bin')
        os.remove(binfile)

        self.assertEqual(data, bindata)

        self.assertRaises(
            Page404Exception,
            self.get, 'download_file', path='../shall_not_pass.txt'
        )

    def test_download_directory(self):
        binfile = os.path.join(self.start, 'testfile.bin')
        bindata = bytes(range(256))

        with open(binfile, 'wb') as f:
            f.write(bindata)
        data = self.get('download_directory', path='start')
        os.remove(binfile)

        iodata = io.BytesIO(data)
        with tarfile.open('start.tgz', mode="r:gz", fileobj=iodata) as tgz:
            tgz_files = [member.name for member in tgz.getmembers() if member.name]
        tgz_files.sort()

        self.assertEqual(tgz_files, ['testfile.bin', 'testfile.txt',])

        self.assertRaises(
            Page404Exception,
            self.get, 'download_directory', path='../../shall_not_pass'
        )

    def test_upload(self):
        c = unichr if PY_LEGACY else chr

        files = {
            'testfile.txt': io.BytesIO(''.join(map(c, range(127))).encode('ascii')),
            'testfile.bin': io.BytesIO(''.join(map(c, range(255))).encode('utf-8')),
        }
        output = self.post('upload',
                           path='upload',
                           data={'file%d' % n: (data, name) for n, (name, data) in enumerate(files.items())}
                           )
        expected_links = sorted(self.url_for('open', path='upload/%s' % i) for i in files)
        self.assertEqual(sorted(output.files), expected_links)
        self.clear(self.upload)


class TestFile(unittest.TestCase):
    def setUp(self):
        self.app = browsepy.app
        self.workbench = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.workbench)

    def testMime(self):
        f = browsepy.File('non_working_path')
        self.assertEqual(f.mimetype, 'application/octet-stream')

        f = browsepy.File('non_working_path_with_ext.txt')
        self.assertEqual(f.mimetype, 'text/plain')


        tmp_txt = os.path.join(self.workbench, 'ascii_text_file')
        with open(tmp_txt, 'w') as f:
            f.write('ascii text')

        # test file command
        f = browsepy.File(tmp_txt)
        self.assertEqual(f.mimetype, 'text/plain; charset=us-ascii')
        self.assertEqual(f.type, 'text/plain')
        self.assertEqual(f.encoding, 'us-ascii')

        # test non-working file command
        bad_path = os.path.join(self.workbench, 'path')
        os.mkdir(bad_path)

        bad_file = os.path.join(bad_path, 'file')
        with open(bad_file, 'w') as f:
            f.write('#!/usr/bin/env bash\nexit 1\n')
        os.chmod(bad_file, os.stat(bad_file).st_mode | stat.S_IEXEC)

        old_path = os.environ['PATH']
        os.environ['PATH'] = bad_path + os.pathsep + old_path

        f = browsepy.File(tmp_txt)
        self.assertEqual(f.mimetype, 'application/octet-stream')

        os.environ['PATH'] = old_path

    def testSize(self):
        test_file = os.path.join(self.workbench, 'test.csv')
        with open(test_file, 'w') as f:
            f.write(',\n'*512)
        f = browsepy.File(test_file)

        default = self.app.config['use_binary_multiples']

        self.app.config['use_binary_multiples'] = True
        self.assertEqual(f.size, '1.00 KiB')

        self.app.config['use_binary_multiples'] = False
        self.assertEqual(f.size, '1.02 KB')

        self.app.config['use_binary_multiples'] = default

        self.assertEqual(f.encoding, 'default')

    def testProperties(self):
        empty_file = os.path.join(self.workbench, 'empty.txt')
        open(empty_file, 'w').close()
        f = browsepy.File(empty_file)

        self.assertEqual(f.basename, 'empty.txt')
        self.assertEqual(f.can_download, True)
        self.assertEqual(f.can_remove, False)
        self.assertEqual(f.can_upload, False)
        self.assertEqual(f.dirname, self.workbench)
        self.assertEqual(f.is_directory, False)


class TestFunctions(unittest.TestCase):
    def test_fmt_size(self):
        fnc = browsepy.fmt_size
        for n, unit in enumerate(browsepy.binary_units):
            self.assertEqual(fnc(2**(10*n)), (1, unit))
        for n, unit in enumerate(browsepy.standard_units):
            self.assertEqual(fnc(1000**n, False), (1, unit))

    def test_empty_iterable(self):
        fnc = browsepy.empty_iterable
        empty, iterable = fnc(i for i in ())
        self.assertTrue(empty)
        self.assertRaises(StopIteration, next, iterable)
        empty, iterable = fnc(i for i in (1, 2))
        self.assertFalse(empty)
        self.assertEqual(tuple(iterable), (1, 2))

    def test_secure_filename(self):
        self.assertEqual(browsepy.secure_filename('/path'), 'path')
        self.assertEqual(browsepy.secure_filename('..'), '')
        self.assertEqual(browsepy.secure_filename('::'), '')
        self.assertEqual(browsepy.secure_filename('\0'), '_')
        self.assertEqual(browsepy.secure_filename('/'), '')
        self.assertEqual(browsepy.secure_filename('C:\\'), '')
        self.assertEqual(browsepy.secure_filename('COM1.asdf', destiny_os='nt'), '')
        self.assertEqual(browsepy.secure_filename('\xf1', fs_encoding='ascii'), '_')

        if PY_LEGACY:
            expected = unicode('\xf1', encoding='latin-1')
            self.assertEqual(browsepy.secure_filename('\xf1', fs_encoding='utf-8'), expected)
            self.assertEqual(browsepy.secure_filename(expected, fs_encoding='utf-8'), expected)
        else:
            self.assertEqual(browsepy.secure_filename('\xf1', fs_encoding='utf-8'), '\xf1')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
browsepy
========

Simple web file browser with directory gzipped tarball download

More details on the `github page <https://github.com/ergoithz/browsepy/>`_.


Development Version
-------------------

The browsepy development version can be installed by cloning the git
repository from `github`_::

    git clone git@github.com:ergoithz/browsepy.git

.. _github: https://github.com/ergoithz/browsepy
"""
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('browsepy/__meta__.py') as f:
    data = {}
    code = compile(f.read(), 'browsepy/__meta__.py', 'exec')
    exec(code, data, data)
    __app__ = data['__app__']
    __version__ = data['__version__']
    __license__ = data['__license__']

with open('README.rst') as f:
    __doc__ = f.read()

setup(
    name=__app__,
    version=__version__,
    url='https://github.com/ergoithz/browsepy',
    download_url = 'https://github.com/ergoithz/browsepy/tarball/0.3.2',
    license=__license__,
    author='Felipe A. Hernandez',
    author_email='ergoithz@gmail.com',
    description='Simple web file browser',
    long_description=__doc__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    keywords = ['web', 'file', 'browser'],
    packages=['browsepy'],
    package_data={'browsepy': ['templates/*', 'static/*']}, # ignored by sdist (see MANIFEST.in), used by bdist_wheel
    install_requires=['flask'],
    test_suite='browsepy.tests',
    zip_safe=False,
    platforms='any'
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import re
import os
import sys
import os.path
import subprocess
import mimetypes
import datetime
import itertools
import tarfile
import shutil
import threading
import codecs

from flask import Flask, Response, request, render_template, redirect, \
                   url_for, send_from_directory, stream_with_context, make_response
from werkzeug.exceptions import NotFound
from werkzeug.utils import cached_property

from .__meta__ import __app__, __version__, __license__, __author__

__basedir__ = os.path.abspath(os.path.dirname(__file__))

app = Flask(__name__,
    static_url_path = os.path.join(__basedir__, "static"),
    template_folder = os.path.join(__basedir__, "templates")
    )
app.config.update(
    directory_base = os.path.abspath(os.getcwd()),
    directory_start = os.path.abspath(os.getcwd()),
    directory_remove = None,
    directory_upload = None,
    directory_tar_buffsize = 262144,
    directory_downloadable = True,
    use_binary_multiples = True,
    )

if "BROWSEPY_SETTINGS" in os.environ:
    app.config.from_envvar("BROWSEPY_SETTINGS")

PY_LEGACY = sys.version_info[0] < 3
if PY_LEGACY:
    FileNotFoundError = type('FileNotFoundError', (OSError,), {})

undescore_replace = '%s:underscore' % __name__
codecs.register_error(undescore_replace,
                      (lambda error: (u'_', error.start + 1))
                      if PY_LEGACY else
                      (lambda error: ('_', error.start + 1))
                      )


class File(object):
    re_mime_validate = re.compile('\w+/\w+(; \w+=[^;]+)*')
    re_charset = re.compile('; charset=(?P<charset>[^;]+)')
    def __init__(self, path):
        self.path = path

    def remove(self):
        if not self.can_remove:
            raise OutsideRemovableBase("File outside removable base")
        if self.is_directory:
            shutil.rmtree(self.path)
        else:
            os.unlink(self.path)

    def download(self):
        if self.is_directory:
            stream = TarFileStream(
                self.path,
                app.config["directory_tar_buffsize"]
                )
            return Response(stream, mimetype="application/octet-stream")
        directory, name = os.path.split(self.path)
        return send_from_directory(directory, name, as_attachment=True)

    @cached_property
    def can_download(self):
        return app.config['directory_downloadable'] or not self.is_directory

    @cached_property
    def can_remove(self):
        dirbase = app.config["directory_remove"]
        if dirbase:
            return self.path.startswith(dirbase + os.sep)
        return False

    @cached_property
    def can_upload(self):
        dirbase = app.config["directory_upload"]
        if self.is_directory and dirbase:
            return dirbase == self.path or self.path.startswith(dirbase + os.sep)
        return False

    @cached_property
    def stats(self):
        return os.stat(self.path)

    _generic_mimetypes = {
        None,
        'application/octet-stream',
        }
    @cached_property
    def mimetype(self):
        mime, encoding = mimetypes.guess_type(self.path)
        mimetype = "%s%s%s" % (mime or "application/octet-stream", "; " if encoding else "", encoding or "")
        if mime in self._generic_mimetypes:
            try:
                output = subprocess.check_output(("file", "-ib", self.path)).decode('utf8').strip()
                if self.re_mime_validate.match(output):
                    # 'file' command can return status zero with invalid output
                    mimetype = output
            except (subprocess.CalledProcessError, FileNotFoundError):
                pass
        return mimetype

    @cached_property
    def is_directory(self):
        return self.type.endswith("directory") or \
               self.type.endswith("symlink") and \
               os.path.isdir(self.path)

    @cached_property
    def parent(self):
        return File(os.path.dirname(self.path))

    @property
    def mtime(self):
        return self.stats.st_mtime

    @property
    def modified(self):
        return datetime.datetime.fromtimestamp(self.mtime).strftime('%Y.%m.%d %H:%M:%S')

    @property
    def size(self):
        size, unit = fmt_size(self.stats.st_size, app.config["use_binary_multiples"])
        if unit == binary_units[0]:
            return "%d %s" % (size, unit)
        return "%.2f %s" % (size, unit)

    @property
    def relpath(self):
        return relativize_path(self.path)

    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def dirname(self):
        return os.path.dirname(self.path)

    @property
    def type(self):
        return self.mimetype.split(";", 1)[0]

    @property
    def encoding(self):
        if ";" in self.mimetype:
            match = self.re_charset.search(self.mimetype)
            gdict = match.groupdict() if match else {}
            return gdict.get("charset") or "default"
        return "default"

    @classmethod
    def listdir_order(cls, path):
        return not os.path.isdir(path), os.path.basename(path).lower()

    def listdir(self):
        pjoin = os.path.join # minimize list comprehension overhead
        content = [pjoin(self.path, i) for i in os.listdir(self.path)]
        content.sort(key=self.listdir_order)
        for i in content:
            yield self.__class__(i)


class TarFileStream(object):
    '''
    Tarfile which compresses while reading for streaming.

    Buffsize can be provided, it must be 512 multiple (the tar block size) for
    compression.
    '''
    def __init__(self, path, buffsize=10240):
        self.path = path
        self.name = os.path.basename(path) + ".tgz"

        self._finished = 0
        self._want = 0
        self._data = bytes()
        self._add = threading.Event()
        self._result = threading.Event()
        self._tarfile = tarfile.open(fileobj=self, mode="w|gz", bufsize=buffsize) # stream write
        self._th = threading.Thread(target=self.fill)
        self._th.start()

    def fill(self):
        self._tarfile.add(self.path, "")
        self._tarfile.close() # force stream flush
        self._finished += 1
        if not self._result.is_set():
            self._result.set()

    def write(self, data):
        self._add.wait()
        self._data += data
        if len(self._data) > self._want:
            self._add.clear()
            self._result.set()
        return len(data)

    def read(self, want=0):
        if self._finished:
            if self._finished == 1:
                self._finished += 1
                return ""
            return EOFError("EOF reached")

        # Thread communication
        self._want = want
        self._add.set()
        self._result.wait()
        self._result.clear()

        if want:
            data = self._data[:want]
            self._data = self._data[want:]
        else:
            data = self._data
            self._data = bytes()
        return data

    def __iter__(self):
        data = self.read()
        while data:
            yield data
            data = self.read()


class OutsideDirectoryBase(Exception):
    pass


class OutsideRemovableBase(Exception):
    pass


binary_units = ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
standard_units = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
def fmt_size(size, binary=True):
    if binary:
        fmt_sizes = binary_units
        fmt_divider = 1024.
    else:
        fmt_sizes = standard_units
        fmt_divider = 1000.
    for fmt in fmt_sizes[:-1]:
        if size < 1000:
            return (size, fmt)
        size /= fmt_divider
    return size, fmt_sizes[-1]

def empty_iterable(iterable):
    try:
        rest = iter(iterable)
        first = next(rest)
        return False, itertools.chain((first,), rest)
    except StopIteration:
        return True, iter(())

def relativize_path(path):
    prefix = os.path.commonprefix((path, app.config["directory_base"]))
    prefix_len = len(prefix)
    if not prefix.endswith(os.sep):
        prefix_len += len(os.sep)
    return path[prefix_len:]

def urlpath_to_abspath(path):
    dirbase = app.config["directory_base"]
    prefix = dirbase
    if not prefix.endswith(os.sep):
        prefix += os.sep
    realpath = os.path.abspath(prefix + path)
    if dirbase == realpath or realpath.startswith(prefix):
        return realpath
    raise OutsideDirectoryBase("%r is not under %r" % (realpath, dirbase))

restricted_names = ('.', '..', '::', os.sep)
restricted_chars = '\/\0'
common_path_separators = '\\/'
nt_device_names = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1', 'LPT2', 'LPT3', 'PRN', 'NUL')
fs_encoding = 'unicode' if os.name == 'nt' else sys.getfilesystemencoding() or 'ascii'
def secure_filename(path, destiny_os=os.name, fs_encoding=fs_encoding):
    '''
    Get rid of parent path components and special filenames.

    If path is invalid or protected, return empty string.

    :param path: unsafe path
    :return: filename or empty string
    '''
    for sep in common_path_separators:
        if sep in path:
            _, path = path.rsplit(sep, 1)

    for character in restricted_chars:
        path = path.replace(character, '_')

    if destiny_os == 'nt':
        fpc = path.split('.', 1)[0].upper()
        if fpc in nt_device_names:
            return ''

    if path in restricted_names:
        return ''

    if fs_encoding != 'unicode':
        if PY_LEGACY and not isinstance(path, unicode):
            path = unicode(path, encoding='latin-1')
        path = path.encode(fs_encoding, errors=undescore_replace).decode(fs_encoding)

    return path


def stream_template(template_name, **context):
    '''
    Some templates can be huge, this function returns an streaming response,
    sending the content in chunks and preventing from timeout.

    Params:
        template_name: template
        **context: parameters for templates.

    Yields:
        HTML strings
    '''
    app.update_template_context(context)
    template = app.jinja_env.get_template(template_name)
    stream = template.generate(context)
    return Response(stream_with_context(stream))


@app.context_processor
def template_globals():
    return {
        'len': len,
        }

@app.route("/browse", defaults={"path":""})
@app.route('/browse/<path:path>')
def browse(path):
    try:
        realpath = urlpath_to_abspath(path)
        directory = File(realpath)
        if directory.is_directory:
            files = directory.listdir()
            empty_files, files = empty_iterable(files)
            return stream_template("browsepy.browse.html",
                dirbase = os.path.basename(app.config["directory_base"]) or '/',
                path = relativize_path(realpath),
                upload = directory.can_upload,
                files = files,
                has_files = not empty_files
                )
    except OutsideDirectoryBase:
        pass
    return NotFound()

@app.route('/open/<path:path>', endpoint="open")
def open_file(path):
    try:
        realpath = urlpath_to_abspath(path)
        if os.path.isfile(realpath):
            return send_from_directory(
                os.path.dirname(realpath),
                os.path.basename(realpath)
                )
    except OutsideDirectoryBase:
        pass
    return NotFound()

@app.route("/download/file/<path:path>")
def download_file(path):
    try:
        realpath = urlpath_to_abspath(path)
        return File(realpath).download()
    except OutsideDirectoryBase:
        pass
    return NotFound()

@app.route("/download/directory/<path:path>.tgz")
def download_directory(path):
    try:
        # Force download whatever is returned
        realpath = urlpath_to_abspath(path)
        return File(realpath).download()
    except OutsideDirectoryBase:
        pass
    return NotFound()

@app.route("/remove/<path:path>", methods=("GET", "POST"))
def remove(path):
    try:
        realpath = urlpath_to_abspath(path)
    except OutsideDirectoryBase:
        return NotFound()
    if request.method == 'GET':
        if not File(realpath).can_remove:
            return NotFound()
        return render_template('browsepy.remove.html',
                               backurl = url_for("browse", path=path).rsplit("/", 1)[0],
                               path = path)
    try:
        f = File(realpath)
        p = f.parent
        f.remove()
    except OutsideRemovableBase:
        return NotFound()
    return redirect(url_for(".browse", path=relativize_path(p.path)))

@app.route("/upload/<path:path>", methods=("POST",))
def upload(path):
    try:
        realpath = urlpath_to_abspath(path)
    except OutsideDirectoryBase:
        return NotFound()

    directory = File(realpath)
    if not directory.is_directory or not directory.can_upload:
        return NotFound()

    for f in request.files.values():
        filename = secure_filename(f.filename)
        if filename:
            f.save(os.path.join(directory.path, filename))
    return redirect(url_for(".browse", path=relativize_path(realpath)))


@app.route("/")
def index():
    try:
        relpath = File(app.config["directory_start"] or
                       app.config["directory_base"]
                       ).relpath
    except OutsideDirectoryBase:
        return NotFound()
    return browse(relpath)

@app.after_request
def page_not_found(response):
    if response.status_code == 404:
        return make_response((render_template('404.html'), 404))
    return response

@app.errorhandler(404)
def page_not_found(e):
    return render_template('404.html'), 404

@app.errorhandler(500)
def internal_server_error(e): # pragma: no cover
    import traceback
    traceback.print_exc()
    return getattr(e, 'message', 'Internal server error'), 500
